package cz.cvut.k36.pr2.hw.hw06;

/**
 * @author Ondrej Hrstka <ondrej.hrstka@agents.fel.cvut.cz>
 */
public interface Pr2List<T> extends Iterable<T> {

    /**
     * Appends element to the end of the list. This means that the element is after all elements
     * in the list, but before the Nil element. If the list is empty,
     * this method is equivalent to the {@link cz.cvut.k36.pr2.hw.hw06.Pr2List#cons(Object)} method.
     * @param t elemnt to be added
     * @return list with added element
     * @throws java.lang.UnsupportedOperationException if the list is view
     */
    public Pr2List<T> append(T t);

    /**
     * Adds element to the beginning of the list.
     * @param t element to be added
     * @return list with added element
     * @throws java.lang.UnsupportedOperationException if the list is view
     */
    public Pr2List<T> cons(T t);

    /**
     * @return {@code true} iff the list is Nil
     */
    public boolean isEmpty();

    /**
     * Filter out all the elements in the list that violate the given predicate.
     * @param predicate to use in filter
     * @return list with filtered elements
     */
    public Pr2List<T> filter(Predicate<T> predicate);

    /**
     * Maps the elements in the list to another elements using the given {@link cz.cvut.k36.pr2
     * .hw.hw06.Function}.
     * @param function
     * @param <O>
     * @return mapped list
     */
    public <O> Pr2List<O> map(Function<T, O> function);

    /**
     * Gets the first element in the list. If the list is empty, the {@link java.util.NoSuchElementException}
     * is thrown.
     * @return the first element in the list
     * @throws java.util.NoSuchElementException if the list is empty
     */
    public T head();

    /**
     * Returns the rest of the list without the first element (which is returned by the {@link cz
     * .cvut.k36.pr2.hw.hw06.Pr2List#head} method. If the list is empty the {@link java.lang
     * .UnsupportedOperationException} is thrown. The returned list can be empty (Nil).
     * @return list without the first element
     * @throws java.lang.UnsupportedOperationException
     * @see Pr2List#head()
     */
    public Pr2List<T> tail();

    public Pr2List<T> view();

    /**
     * @return the number of elements in the list
     */
    public int size();
}
